@@ -11,7 +11,7 @@ use Sys::Hostname 'hostname';
 
 has 'mysql';
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 
 sub dequeue {
   my ($self, $id, $wait, $options) = @_;
@@ -40,18 +40,40 @@ sub enqueue {
 
   my $seconds = $db->dbh->quote($options->{delay} // 0);
   $db->query(
-    "insert into minion_jobs (`args`, `attempts`, `delayed`, `priority`, `queue`, `task`)
-     values (?, ?, (DATE_ADD(NOW(), INTERVAL $seconds SECOND)), ?, ?, ?)",
+    "insert into minion_jobs (`args`, `attempts`, `delayed`, `priority`, `queue`, `task`, `notes`)
+     values (?, ?, (DATE_ADD(NOW(), INTERVAL $seconds SECOND)), ?, ?, ?, ?)",
      encode_json($args), $options->{attempts} // 1,
-     $options->{priority} // 0, $options->{queue} // 'default', $task
+     $options->{priority} // 0, $options->{queue} // 'default', $task,
+     encode_json( $options->{notes} // {} ),
   );
   my $job_id = $db->dbh->{mysql_insertid};
 
+  if ( my @parents = @{ $options->{parents} || [] } ) {
+    $db->query(
+      "INSERT IGNORE INTO minion_jobs_depends (`parent_id`, `child_id`) VALUES "
+      . join( ", ", map "( ?, ? )", @parents ),
+      map { $_, $job_id  } @parents
+    );
+  }
+
   $self->mysql->pubsub->notify("minion.job" => $job_id);
 
   return $job_id;
 }
 
+sub note {
+  my ($self, $id, $key, $value) = @_;
+  my $job = $self->mysql->db->query(
+    'SELECT notes FROM minion_jobs WHERE id=?', $id,
+  )->hash || return 0;
+  my $notes = decode_json( $job->{notes} );
+  $notes->{ $key } = $value;
+  return !!$self->mysql->db->query(
+    'UPDATE minion_jobs SET notes = ? WHERE id = ?',
+    encode_json( $notes ), $id,
+  )->rows;
+}
+
 sub fail_job   { shift->_update(1, @_) }
 sub finish_job { shift->_update(0, @_) }
 
@@ -64,6 +86,11 @@ sub list_jobs {
     push @where, 'state in (' . join( ',', ('?') x @states ) . ')';
     push @params, @states;
   }
+  if ( my $queue = $options->{queue} ) {
+    my @queues = ref $queue eq 'ARRAY' ? @$queue : ( $queue );
+    push @where, 'queue in (' . join( ',', ('?') x @queues ) . ')';
+    push @params, @queues;
+  }
   if ( my $task = $options->{task} ) {
     my @tasks = ref $task eq 'ARRAY' ? @$task : ( $task );
     push @where, 'task in (' . join( ',', ('?') x @tasks ) . ')';
@@ -76,7 +103,6 @@ sub list_jobs {
   }
 
   my $where = @where ? 'WHERE ' . join( ' AND ', @where ) : '';
-  # XXX: notes column missing
   my $jobs = $self->mysql->db->query(
     "SELECT
       id, args, attempts,
@@ -85,17 +111,26 @@ sub list_jobs {
       UNIX_TIMESTAMP(finished) AS finished, priority,
       queue, result, UNIX_TIMESTAMP(retried) AS retried, retries,
       UNIX_TIMESTAMP(started) AS started, state, task,
-      worker
+      GROUP_CONCAT( child_jobs.child_id SEPARATOR ':' ) AS children,
+      GROUP_CONCAT( parent_jobs.parent_id SEPARATOR ':' ) AS parents,
+      worker, notes
     FROM minion_jobs
+    LEFT JOIN minion_jobs_depends child_jobs ON minion_jobs.id=child_jobs.parent_id
+    LEFT JOIN minion_jobs_depends parent_jobs ON minion_jobs.id=parent_jobs.child_id
     $where
+    GROUP BY minion_jobs.id, child_jobs.parent_id, parent_jobs.child_id
     ORDER BY id DESC
     LIMIT ?
     OFFSET ?", @params, $limit, $offset,
   )->hashes;
-  $jobs->map( _decode_json_fields(qw{ args result }) )
-    # XXX: Job children/parents not supported
-    # Add fake arrayrefs to make the Minion UI work
-    ->map( sub { $_[0]->{parents} = $_[0]->{children} = []; return $_[0] } );
+  $jobs->map( _decode_json_fields(qw{ args result notes }) )
+    ->each( sub {
+      $_->{children} = [ split /:/, $_->{children} // '' ];
+      $_->{parents} = [ split /:/, $_->{parents} // '' ];
+    } );
+
+  #; use Data::Dumper;
+  #; say Dumper $jobs;
 
   my $total = $self->mysql->db->query(
     'SELECT COUNT(*) AS count FROM minion_jobs',
@@ -129,10 +164,9 @@ sub list_workers {
   }
 
   my $where = @where ? 'WHERE ' . join ' AND ', @where : '';
-  # XXX: Missing "status" column for workers
   my $sql = "SELECT
     id, UNIX_TIMESTAMP(notified) AS notified, host, pid,
-    UNIX_TIMESTAMP(started) AS started
+    UNIX_TIMESTAMP(started) AS started, status
   FROM minion_workers $where ORDER BY id DESC LIMIT ? OFFSET ?";
   my $workers = $self->mysql->db->query($sql, @params, $limit, $offset)
     ->hashes;
@@ -140,6 +174,7 @@ sub list_workers {
   # Add jobs to each worker
   my $jobs_sql = q{SELECT id FROM minion_jobs WHERE state='active' AND worker=?};
   $workers->map( sub {
+      $_->{status} = decode_json( $_->{status} );
       $_->{jobs} = $self->mysql->db->query($jobs_sql, $_->{id})->arrays->flatten->to_array
   } );
 
@@ -166,20 +201,15 @@ sub new {
 }
 
 sub register_worker {
-  my ($self, $id) = @_;
-
-  ### TODO: is this the same business logic?
-  if ($id) {
-      my $sql
-        = 'update minion_workers set notified = now() where id = ?';
-      return $id if $self->mysql->db->query($sql, $id)->{affected_rows};
-  }
+  my ($self, $id, $options) = @_;
 
   my $db = $self->mysql->db;
-  my $sql = 'insert into minion_workers (host, pid) values (?, ?)';
-  $db->query($sql, hostname, $$);
+  my $sql = q{INSERT INTO minion_workers (id, host, pid, status)
+    VALUES (?, ?, ?, ?)
+    ON DUPLICATE KEY UPDATE notified=NOW(), host=VALUES(host), pid=VALUES(pid), status=VALUES(status)};
+  $db->query($sql, $id, hostname, $$, encode_json( $options->{status} // {} ) );
 
-  return $db->dbh->{mysql_insertid};
+  return $id // $db->dbh->{mysql_insertid};
 }
 
 sub remove_job {
@@ -198,34 +228,57 @@ sub repair {
   my $minion = $self->minion;
   $db->query(
     "delete from minion_workers
-     where notified < (DATE_ADD(NOW(), INTERVAL ? SECOND))", $minion->missing_after
+     where notified < (DATE_SUB(NOW(), INTERVAL ? SECOND))",
+     $minion->missing_after
   );
 
   # Abandoned jobs
-  $db->query(
-    "update minion_jobs as j
-     set finished = now(), result = ?,
-       state = 'failed'
+  my $fail = $db->query(
+    "select id, retries from minion_jobs as j
      where state = 'active'
-       and not exists(select 1 from minion_workers where id = j.worker)",
-   encode_json('Worker went away')
+       and not exists (select 1 from minion_workers where id = j.worker)"
+  )->hashes;
+  $fail->each(sub { $self->fail_job(@$_{qw(id retries)}, 'Worker went away') });
+
+  # Old jobs with no unresolved dependencies
+  $db->query( q{
+    DELETE FROM minion_jobs
+    WHERE state = 'finished'
+      AND finished <= (DATE_SUB(NOW(), INTERVAL ? SECOND))
+      AND state='finished'
+      AND NOT EXISTS (
+        SELECT 1 FROM ( SELECT id, state FROM minion_jobs ) AS child
+        LEFT JOIN minion_jobs_depends depends ON child.id=depends.child_id
+        WHERE parent_id=minion_jobs.id AND child.state != 'finished'
+      )
+    }, $minion->remove_after,
   );
 
-  # Old jobs
-  $db->query(
-    "delete from minion_jobs
-     where state = 'finished' and finished < (DATE_SUB(NOW(), INTERVAL ? SECOND))",
-    $minion->remove_after
-  );
 }
 
 sub reset {
     my $self = shift;
 
-    $self->mysql->db->query("truncate table minion_jobs");
+    $self->mysql->db->query("delete from minion_jobs");
+    $self->mysql->db->query("truncate table minion_locks");
     $self->mysql->db->query("truncate table minion_workers");
 }
 
+sub lock {
+  my ($self, $name, $duration, $options) = (shift, shift, shift, shift // {});
+  return !!$self->mysql->db->query('SELECT minion_lock(?, ?, ?)',
+    $name, $duration, $options->{limit} || 1)->array->[0];
+}
+
+sub unlock {
+  !!shift->mysql->db->query(
+    'DELETE FROM minion_locks WHERE id IN (
+       SELECT b.id FROM (SELECT id, expires, name FROM minion_locks) b
+       WHERE b.expires > NOW() AND b.name = ? ORDER BY EXPIRES
+     ) LIMIT 1', shift
+  )->rows;
+}
+
 sub retry_job {
   my ($self, $id, $retries) = (shift, shift, shift);
   my $options = shift // {};
@@ -233,12 +286,13 @@ sub retry_job {
   my $seconds = $self->mysql->db->dbh->quote($options->{delay} // 0);
 
   return !!$self->mysql->db->query(
-    "update `minion_jobs`
-     set `finished` = null, priority = coalesce(?, priority),
-      `queue` = coalesce(?, queue), `result` = null, `retried` = now(),
-       `retries` = retries + 1, `started` = null, `state` = 'inactive',
-       `worker` = null, `delayed` = (DATE_ADD(NOW(), INTERVAL $seconds SECOND))
-     where `id` = ? and retries = ? and `state` in ('failed', 'finished')",
+    "UPDATE `minion_jobs`
+     SET attempts = COALESCE(?, attempts),
+       `delayed` = DATE_ADD(NOW(), INTERVAL $seconds SECOND),
+       priority = COALESCE(?, priority), queue = COALESCE(?, queue),
+       retried = NOW(), retries = retries + 1, state = 'inactive'
+     WHERE id = ? AND retries = ?",
+     $options->{attempts},
      @$options{qw(priority queue)}, $id, $retries
   )->{affected_rows};
 }
@@ -264,6 +318,14 @@ sub stats {
 
   my $uptime = $db->query( 'SHOW GLOBAL STATUS LIKE "Uptime"' )->hash->{Value};
 
+  $sql = q{
+    SELECT
+      SUM(CASE WHEN `state` = "inactive" AND `delayed` > NOW() THEN 1 ELSE 0 END) AS delayed_jobs,
+      COUNT(*) AS enqueued_jobs
+      FROM minion_jobs
+    };
+  %$states = ( %$states, %{ $db->query($sql)->hash } );
+
   return {
     active_workers   => $active,
     inactive_workers => $all - $active,
@@ -271,6 +333,8 @@ sub stats {
     inactive_jobs    => $states->{inactive} || 0,
     failed_jobs      => $states->{failed} || 0,
     finished_jobs    => $states->{finished} || 0,
+    enqueued_jobs    => $states->{enqueued_jobs} || 0,
+    delayed_jobs     => $states->{delayed_jobs} || 0,
     uptime           => $uptime || 0,
   };
 }
@@ -292,13 +356,25 @@ sub _try {
   my $db = $self->mysql->db;
 
   my $tx = $db->begin;
-  my $job = $tx->db->query(qq(select id, args, retries, task from minion_jobs
-    where state = 'inactive' and `delayed` <= NOW() and queue in ($qq)
-    and task in ($qt)
-    order by priority desc, created limit 1 for update),
+  my $job = $tx->db->query(qq{
+    SELECT job.id, job.args, job.retries, job.task
+    FROM minion_jobs job
+    WHERE job.state = 'inactive' AND job.`delayed` <= NOW()
+      AND NOT EXISTS (
+        SELECT 1 FROM minion_jobs_depends depends
+        LEFT JOIN minion_jobs parent ON parent.id=depends.parent_id
+        WHERE child_id=job.id AND parent.id=depends.parent_id AND parent.state IN ( 'inactive', 'active', 'failed' )
+      )
+      AND job.queue IN ($qq) AND job.task IN ($qt)
+    GROUP BY job.id
+    ORDER BY job.priority DESC, job.created
+    LIMIT 1 FOR UPDATE},
    @{ $options->{queues} || ['default']}, @{ $tasks }
   )->hash;
 
+  #; use Data::Dumper;
+  #; say "Dequeuing job: " . Dumper $job;
+
   return undef unless $job;
 
   $tx->db->query(
@@ -309,18 +385,21 @@ sub _try {
 
   $job->{args} = $job->{args} ? decode_json($job->{args}) : undef;
 
-  $job;
+  return $job;
 }
 
 sub _update {
   my ($self, $fail, $id, $retries, $result) = @_;
-  return undef unless $self->mysql->db->query(
+  my $updated = $self->mysql->db->query(
     "update minion_jobs
      set finished = now(), result = ?, state = ?
      where id = ? and retries = ? and state = 'active'",
      encode_json($result), $fail ? 'failed' : 'finished', $id,
     $retries
   )->{affected_rows};
+  #; say "Updated $updated job rows (id: $id, fail: $fail, result: @{[encode_json( $result )]})";
+  return undef unless $updated;
+
   my $job = $self->list_jobs( 0, 1, { ids => [$id] } )->{jobs}[0];
   return 1 if !$fail || (my $attempts = $job->{attempts}) == 1;
   return 1 if $retries >= ( $attempts - 1 );
@@ -776,7 +855,7 @@ Minion::Backend::mysql
 
 =head1 VERSION
 
-version 0.11
+version 0.12
 
 =head1 SYNOPSIS
 
@@ -1275,3 +1354,43 @@ ALTER TABLE minion_jobs ADD COLUMN attempts INT NOT NULL DEFAULT 1;
 ALTER TABLE minion_jobs MODIFY COLUMN args MEDIUMBLOB NOT NULL;
 ALTER TABLE minion_jobs MODIFY COLUMN result MEDIUMBLOB;
 
+-- 6 up
+ALTER TABLE minion_workers ADD COLUMN status MEDIUMBLOB;
+ALTER TABLE minion_jobs ADD COLUMN notes MEDIUMBLOB;
+CREATE TABLE IF NOT EXISTS minion_locks (
+  id      SERIAL NOT NULL PRIMARY KEY,
+  name    VARCHAR(255) NOT NULL,
+  expires TIMESTAMP NOT NULL,
+  INDEX (name, expires)
+);
+DELIMITER //
+CREATE FUNCTION minion_lock( $1 VARCHAR(255), $2 INTEGER, $3 INTEGER) RETURNS BOOL
+BEGIN
+  DECLARE new_expires TIMESTAMP DEFAULT DATE_ADD( NOW(), INTERVAL 1*$2 SECOND );
+  DELETE FROM minion_locks WHERE expires < NOW();
+  IF (SELECT COUNT(*) >= $3 FROM minion_locks WHERE name = $1)
+  THEN
+    RETURN FALSE;
+  END IF;
+  IF new_expires > NOW()
+  THEN
+    INSERT INTO minion_locks (name, expires) VALUES ($1, new_expires);
+  END IF;
+  RETURN TRUE;
+END
+//
+DELIMITER ;
+CREATE TABLE minion_jobs_depends (
+  parent_id BIGINT UNSIGNED NOT NULL,
+  child_id BIGINT UNSIGNED NOT NULL,
+  FOREIGN KEY (parent_id) REFERENCES minion_jobs(id) ON DELETE CASCADE,
+  FOREIGN KEY (child_id) REFERENCES minion_jobs(id) ON DELETE CASCADE
+);
+
+-- 6 down
+ALTER TABLE minion_workers DROP COLUMN status;
+ALTER TABLE minion_jobs DROP COLUMN notes;
+DROP TABLE IF EXISTS minion_locks;
+DROP FUNCTION IF EXISTS minion_lock;
+DROP TABLE minion_jobs_depends;
+
